#############################################################################################
# Knox meets Cox example
#
# [Authors, 2008]
# created   3 May 2008
# altered  27 Feb 2009
#
# Modification of Knox permutation test to account for changing risk sets
#   and effects of space-time interactions in local covariates
#
# With the initial random seed =850, nsim =10000, and other settings at default values,
# this program will reproduce the results in the paper. 
#############################################################################################

graphics.off()
library(survival)
set.seed(850)

############################################################################
# CHANGE THIS DIRECTORY NAME FOR YOUR FILE SYSTEM, TO INDICATE WHERE
# THE INPUT FILES ARE LOCATED
#
# USE FORWARD SLASHES TO SEPARATE SUBDIRECTORIES
# END THE STRING WITH A FORWARD SLASH
############################################################################

dir <- "Q:/example/"

setwd(dir)

############################################################################
# CHANGE THESE PARAMETERS TO ALTER THE DEFINITIONS OF CLOSE IN TIME, 
# CLOSE IN SPACE, OR EARLY/LATE TRANSITIONS
#
# FOR EXPLORATORY EXPERIMENTS WE RECOMMEND CHANGING nsim TO 200, 
#   RATHER THAN 10000
############################################################################

time.cutoff  <- 0        # default = 0
km.cutoff    <- 100      # default = 100
lag.cutoff   <- 1        # default = 1
space.rule   <- "mixed"  # default = "mixed". Choose "km", "lag", or "mixed"
EL.cutoff    <- .60      # default = .60
nsim         <- 10000    # default = 10000

SAVE.RESULTS <- FALSE    # if TRUE, program will write the results to a .dput file

###########################################################################
# DO NOT CHANGE ANYTHING FROM HERE ON...
###########################################################################

nreg        <- 502
ntime       <- 6
n           <- 502
Npair       <- n*(n-1)/2


#--------------------------------------------------------------
# read data for Brazilian fertility transitions and covariates
#
# read the geographic information about distances between the 502 regions
# distances are in 2 forms: km between centroids, and lags (shortest number
# of cross-border moves to get from one region to another -- self is 0,
#  neighbors are 1, neighbors of neighbors are 2, ...)
#
# distances were precomputed and stored in R's dput format,
# as 502 x 502 matrices
#--------------------------------------------------------------


event.data <- read.csv("brazil_transition_times_and_covariates.csv")
Dkm        <- dget("brazil_regions_km_distances.dput")
Dlag       <- dget("brazil_regions_lag_distances.dput")

#--------------------------------------------------------------
# construct the event map, with events in chronological order
#  in R the map will be a list with 2 components:
#    s = an n-vector of event locations (i.e, region numbers)
#    t = an n-vector of event times 
#         (1=pre1960,2=1960-1970,...,5=1991-2000,6=post2000)
#--------------------------------------------------------------
chron <- order(event.data[,"time"])
M     <- list(s=event.data[chron,"region"],t=event.data[chron,"time"])

#print the map
M

#----------------------------------------------
# newmap(perm)
# generates a permutation map
# by reordering event locations
#----------------------------------------------
newmap <- function(map,perm) {  list(s=map$s[perm],t=map$t) }


#----------------------------------------------
# ij.pairs(n)
# this function uses an R trick to generate the (i,j) indices for each
# distinct pair of event numbers 1...n. 
# The result is a [(n)(n-1)/2] x 2 matrix of index numbers
# 
# for example, if pairs <- ij.pairs(5)
#  then pairs is a 10 x 2 matrix
#    [ 1  2 ]
#    [ 1  3 ]
#    [ 1  4 ]
#    [ 1  5 ]
#    [ 2  3 ]
#    [ 2  4 ]
#    [ 2  5 ]
#    [ 3  4 ]
#    [ 3  5 ]
#    [ 4  5 ]
#
#  pairs[2,] is the 2nd distinct combination of 1..5: (1,3) 
#  pairs[6,] is the 6th distinct combination of 1..5: (2,4) 
#  and so forth
#----------------------------------------------

ij.pairs <- function(n) {
  A  <- matrix(0,n,n)
  ii <- col(A)[row(A)>col(A)]
  jj <- row(A)[row(A)>col(A)]
  B  <- matrix(c(ii,jj),n*(n-1)/2,2,dimnames=list(NULL,c("i","j")))
  return(B)
} # ij.pairs

pairs <- ij.pairs(n)

#-------------------------------------------------------
# construct an Npair-vector of T/F dummies, indicating
# whether each pair of events in the map is close in time
# under the current definition
#
# because we will leave these times unchanged in every permutation
# map, the close.in.time vector only needs to be calculated once
#-------------------------------------------------------

close.in.time <- ( abs(M$t[pairs[,"i"]] - M$t[pairs[,"j"]]) <= time.cutoff )

#Nt is the number of pairs close in time
Nt <- sum(close.in.time)

print(list(Nt=Nt))

#----------------------------------------------
# close.in.space(perm)
# a function to construct an Npair-vector of T/F dummies, indicating
# whether each pair of events in the permutation map is 
# close in space under the current definition
#
# unlike close.in.time, this vector is different for every
# map and needs to be computed many times
#----------------------------------------------

close.in.space <- function(perm=1:n) {
  Mstar    <- newmap(M,perm)
  km.dist  <- Dkm[Mstar$s,Mstar$s][pairs]
  lag.dist <- Dlag[Mstar$s,Mstar$s][pairs]

  vec <- switch(space.rule,
           lag   = (lag.dist <= lag.cutoff),
           km    = (km.dist  <= km.cutoff),
           mixed = (lag.dist <= lag.cutoff | km.dist  <= km.cutoff),
           "bad space rule")

  return(vec)
} # close.in.space


Ns <- sum( close.in.space() )

print(list(Ns=Ns))

#-------------------------------------------------------------------
# function to calculate the Knox statistic for a permutation map
#  returns a list with 2 components:
#     X = Knox stat
#     tab = contingency table for (close in space)x(close in time)
#-------------------------------------------------------------------
Knox <- function(perm=1:n) {
  cis <- close.in.space(perm)
  tab <- table(cis,close.in.time)
  tab <- cbind(tab,apply(tab,1,sum)) #add right margin
  tab <- rbind(tab,apply(tab,2,sum)) #add bottom margin
  X   <- sum(cis & close.in.time)
  return(list(X=X,tab=tab))
} 

Xobs <- Knox()$X
print(list(Knox.for.observed.map=Knox()))

##############################################################################
# MAIN PROGRAM
##############################################################################

#--------------------------------------------------------------------
# arrange data for Cox PH regression with time-dependent covariates
# this requires one observation for each region-period of risk
# In the Brazilian data this is slightly complicated by the fact that
#  data are missing for 107 regions in 1960, meaning that they were not
#  at risk of a transition in the pre-1960 period
#
# the strategy here is to first create a full set of 2510 (502x5) region by period 
# observations, and then eliminate (region,pd) combinations not in the risk set
#--------------------------------------------------------------------

start <- matrix(0:4,502,5,byrow=T)
stop  <- start+1
event <- (M$t == stop)

# arrange the 5 covariates into a 2510 x 5 matrix
Zdata <- as.numeric(as.matrix(event.data)[chron,7:31])
Z <- matrix(Zdata,ncol=5,
      dimnames=list(NULL,c("elec","prim","flfp","educ","cath")))

# identify which region-periods are not in the risk set
# because of missing 1960 data
no.data <- is.na(Z[,"elec"])

# identify which region-periods are not in the risk set
# because a transition had already happened
event.already <- ( col(event) > M$t )

# at.risk will be a 2510 x 1 vector of T/F, indicating which
# region-periods to include in the PH regression

at.risk <- !no.data & !event.already

Ydata <- Surv(start[at.risk],stop[at.risk],event[at.risk])
Xdata <- Z[at.risk,]
mod   <- coxph(Ydata ~ Xdata, x=T,y=T)

summary(mod)

# use the model coefs to construct a 502 x 5 matrix of predicted 
# relative probabilities. These will be used to sample permutations 
# non-uniformly in the modified Knox test. Probabilities for the 
# region-periods with missing data will be zero.

# Zbeta is a 2510 x 1 vector of linear predictors from the Cox PH model
Zbeta          <-  Z %*% mod$coefficients  
Zbeta[no.data] <- -Inf
prob           <- matrix(exp(Zbeta),nreg,5)


#-----------------------------
# function to draw an event order
# matching the observed times 
# (for this data set, 29 t=1 transitions, 56 t=2, etc
#-----------------------------

nevent <- table(M$t)   # vector of event counts per period

gen.perm <- function(pmat) {
  P   <- pmat  # temporary copy of 502 x 5 (region x period) probs
  ord <- NULL

  for (pd in 1:5) {
    # select regions with simulated transition in this period
    trans <- sample(502,nevent[pd],prob=P[,pd]) 
   
    # add selected regions to the permutation
    ord <- c(ord,trans)

    # zero out probs for the selected regions (they leave risk set)
    P[trans,] <- 0
  }
  
  #add unselected units (2 in this data set) to the end of ord
  ord <- c(ord,setdiff(1:nreg,ord))
  return(ord)  
}

#--------------------------------------------------
# plotting functions for P(Xstar >= Xobs | H0)
#--------------------------------------------------
tail <- function(X) {
  val  <- sort(unique(X))
  phat <- apply(outer(X,val,">="),2,sum) / length(X)
  return(list(x=val,y=phat))
}

signif.plot <- function(x.std,x.mod,x.obs,include.mod=T) {
    n  <- length(x.std)
    xx <- range(x.std,x.mod,x.obs)
    plot(tail(x.std),type="s",xlim=xx,
     xlab="Knox Stat X",ylab="Signif Level p0(X)",bty="l")

    abline(h=.05,col="grey",lwd=2)
    lines(tail(x.std),type="s")
    if (include.mod) lines(tail(x.mod),type="s",col="red",lwd=3)
    text(1.05*xx[1],.07,"p=.05",col="grey")
    abline(v=x.obs)
    text(.97*x.obs,.50,paste("X(M)=",x.obs,sep=""))
    text(620,.50,"Standard")
    if (include.mod) text(770,.50,"Modified\n(with covariates)",col="red")
}


#--------------------------------------------------
#  Knox simulation
#    keep track of Knox stats for permutations
#    and simulated transition times for each region
#--------------------------------------------------

# region x period simulated transition times
sim.count <- matrix(0,nreg,6)   

Xstar.std <- rep(NA,nsim)
Xstar.mod <- rep(NA,nsim)

for (sim in 1:nsim) {
  #standard (uniform draws for permutations)
  ord            <- sample(nreg)
  Xstar.std[sim] <- Knox(ord)$X

  #modified (non-uniform draws for permutations based on Cox PH probs)
  ord            <- gen.perm(prob)
  Xstar.mod[sim] <- Knox(ord)$X

  #indices for regions and simulation times in modified simulation
  ix <- cbind(ord,M$t)

  # add 1 to the counts for those indices
  sim.count[ix] <- sim.count[ix] + 1

  if (sim %% 100 == 0) {
     signif.plot(Xstar.std[1:sim],Xstar.mod[1:sim],Xobs)
     ti <- paste("Close in time if t difference LE",time.cutoff,"\n",
                switch(space.rule,
                   "km"    = paste("Close in space if dist LE",km.cutoff,"km"),
                   "mixed" = paste("Close in space if dist LE",km.cutoff,"km or lag dist LE",lag.cutoff),
                   "lag"   = paste("Close in space if lag dist LE",lag.cutoff)
                       ),"\n",
     sim,"/",nsim,"simulations complete")
     title(ti)
  }
}

## cumulative distribution plots
signif.plot(Xstar.std,Xstar.mod,Xobs)

# print summary stats for Knox null distributions

print(list(std.quantiles=quantile(Xstar.std,c(0,.50,.90,.95,.99,1)),std.mean=mean(Xstar.std)))
print(list(mod.quantiles=quantile(Xstar.mod,c(0,.50,.90,.95,.99,1)),mod.mean=mean(Xstar.mod)))

#---------------------------------------------------
# use some R tricks to quickly count early and late transitions
#---------------------------------------------------

# proportion of simulated transitions occuring before and after observed
before <- apply(sim.count * (col(sim.count) < M$t), 1, sum) / nsim
after  <- apply(sim.count * (col(sim.count) > M$t), 1, sum) / nsim

# nreg x 1 vectors of T/F dummies for early or late transitions
E <- (after  > EL.cutoff)
L <- (before > EL.cutoff)

print("-------------- Early and Late Transitions, by intercensal period -------------")
table(E,M$t)
table(L,M$t)

#--------------------------------------------------------------------------------------
# SAVE THE OUTPUT FROM THIS RUN FOR LATER MAPPING, TABULATION, etc.
#   the selected regions are the examples used in the paper's tables
#--------------------------------------------------------------------------------------

LL <- list(nreg=nreg, ntime=ntime, n=n, Npair=Npair, time.cutoff=time.cutoff, km.cutoff=km.cutoff, lag.cutoff=lag.cutoff,
          space.rule=space.rule, EL.cutoff=EL.cutoff, nsim=nsim, E=E, L=L, 
          selected.regions = match(c(1,24,49,97,130,322,383,460),M$s),
          Xstar.std=Xstar.std, Xstar.mod=Xstar.mod, sim.count=sim.count, map=M, Knox.observed=Knox(), model=mod, prob=prob)

if (SAVE.RESULTS) dput(LL,paste(dir,"knox_results.dput",sep=""))













